"""
MIT License

Copyright (c) 2019-Present PythonistaGuild

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from __future__ import annotations

import asyncio
import logging
from typing import TYPE_CHECKING, Any, Optional

import aiohttp

import wavelink

from . import __version__
from .backoff import Backoff
from .enums import NodeStatus, TrackEventType
from .exceptions import *
from .payloads import TrackEventPayload

if TYPE_CHECKING:
    from .node import Node
    from .player import Player


logger: logging.Logger = logging.getLogger(__name__)


class Websocket:

    __slots__ = (
        'node',
        'socket',
        'retries',
        'retry',
        '_original_attempts',
        'backoff',
        '_listener_task'
    )

    def __init__(self, *, node: Node) -> None:
        self.node: Node = node
        self.socket: aiohttp.ClientWebSocketResponse | None = None

        self.retries: int | None = node._retries
        self.retry: float = 1
        self._original_attempts: int | None = node._retries

        self.backoff: Backoff = Backoff()

        self._listener_task: asyncio.Task | None = None

    @property
    def headers(self) -> dict[str, str]:
        assert self.node.client is not None
        assert self.node.client.user is not None

        return {
            'Authorization': self.node.password,
            'User-Id': str(self.node.client.user.id),
            'Client-Name': f'Wavelink/{__version__}'
        }

    def is_connected(self) -> bool:
        return self.socket is not None and not self.socket.closed

    async def connect(self) -> None:
        if self.node.status is NodeStatus.CONNECTED:
            logger.error(f'The Node <{self.node!r}> is already in a connected state. Disregarding.')
            return

        self.node._status = NodeStatus.CONNECTING

        try:
            self._listener_task.cancel()
        except Exception as e:
            logger.debug(f'An error was raised while cancelling the websocket listener. {e}')

        uri: str = self.node._host
        heartbeat: float = self.node.heartbeat

        try:
            self.socket = await self.node._session.ws_connect(url=uri, heartbeat=heartbeat, headers=self.headers)
        except Exception as e:
            if isinstance(e, aiohttp.WSServerHandshakeError) and e.status == 401:
                raise AuthorizationFailed from e
            else:
                logger.error(f'An error occurred connecting to node: "{self.node}". {e}')

        if self.is_connected():
            self.retries = self._original_attempts
            # TODO - Configure Resuming...
        else:
            await self._reconnect()
            return

        self._listener_task = asyncio.create_task(self._listen())

    async def _reconnect(self) -> None:
        self.node._status = NodeStatus.CONNECTING
        self.retry = self.backoff.calculate()

        if self.retries == 0:
            logger.error('Wavelink 2.0 was unable to connect, and has exhausted the reconnection attempt limit. '
                         'Please check your Lavalink Node is started and your connection details are correct.')

            await self.cleanup()
            return

        retries = f'{self.retries} attempt(s) remaining.' if self.retries else ''
        logger.error(f'Wavelink 2.0 was unable to connect, retrying connection in: "{self.retry}" seconds. {retries}')

        if self.retries:
            self.retries -= 1

        await asyncio.sleep(self.retry)
        await self.connect()

    async def _listen(self) -> None:
        while True:
            message = await self.socket.receive()

            if message.type in (aiohttp.WSMsgType.CLOSED, aiohttp.WSMsgType.CLOSING):

                for player in self.node.players.copy().values():
                    await player._update_event(data=None)

                asyncio.create_task(self._reconnect())
                return

            if message.data == 1011:
                logger.error('Lavalink encountered an internal error which can not be resolved. '
                             'Make sure your Lavalink sever is up to date, and try restarting.')

                await self.cleanup()
                return

            if message.data is None:
                logger.info('Received a message from Lavalink with empty data. Disregarding.')
                continue

            data = message.json()
            logger.debug(f'Received a message from Lavalink: {data}')

            op = data.get('op', None)
            if not op:
                logger.info('Message "op" from Lavalink was None. Disregarding.')
                continue

            if op == 'ready':
                self.node._status = NodeStatus.CONNECTED
                self.node._session_id = data['sessionId']

                self.dispatch('node_ready', self.node)

            elif op == 'stats':
                payload = ...
                logger.debug(f'Stats Update: {data}')
                self.dispatch('stats_update', data)

            elif op == 'event':
                logger.debug(f'Websocket Event: {data}')
                player = self.get_player(data)

                if player is None:
                    logger.debug('Received payload from Lavalink without an attached player. Disregarding.')
                    continue

                if data['type'] == 'WebSocketClosedEvent':
                    if data['code'] == 4014:
                        continue

                track = await self.node.build_track(cls=wavelink.GenericTrack, encoded=data['encodedTrack'])
                payload: TrackEventPayload = TrackEventPayload(data=data, track=track, player=player)

                if payload.event is TrackEventType.END and payload.reason != 'REPLACED':
                    player._source = None

                self.dispatch('track_event', payload)

                if payload.event is TrackEventType.END:
                    self.dispatch('track_end', payload)
                    asyncio.create_task(player._auto_play_event(payload))

                elif payload.event is TrackEventType.START:
                    self.dispatch('track_start', payload)

            elif op == 'playerUpdate':
                player = self.get_player(data)
                if player is None:
                    logger.debug('Received payload from Lavalink without an attached player. Disregarding.')
                    continue

                await player._update_event(data)
                self.dispatch("player_update", data)
                logger.debug(f'Websocket Player Update: {data}')

            else:
                logger.info(f'Received unknown payload from Lavalink: <{data}>. '
                            f'If this continues consider making a ticket on the Wavelink GitHub. '
                            f'https://github.com/PythonistaGuild/Wavelink')

    def get_player(self, payload: dict[str, Any]) -> Optional['Player']:
        return self.node.players.get(int(payload['guildId']), None)

    def dispatch(self, event, *args: Any, **kwargs: Any) -> None:
        self.node.client.dispatch(f"wavelink_{event}", *args, **kwargs)

    # noinspection PyBroadException
    async def cleanup(self) -> None:
        try:
            await self.socket.close()
        except AttributeError:
            pass

        try:
            self._listener_task.cancel()
        except Exception:
            pass

        self.node._status = NodeStatus.DISCONNECTED



"""
MIT License

Copyright (c) 2019-Present PythonistaGuild

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from __future__ import annotations

import datetime
import logging
from typing import TYPE_CHECKING, Any, Union

import discord
from discord.utils import MISSING

from .enums import *
from .ext import spotify
from .filters import Filter
from .node import Node, NodePool
from .payloads import TrackEventPayload
from .queue import Queue
from .tracks import *


if TYPE_CHECKING:
    from discord.types.voice import GuildVoiceState, VoiceServerUpdate
    from typing_extensions import Self

    from .types.events import PlayerState, PlayerUpdateOp
    from .types.request import EncodedTrackRequest, Request
    from .types.state import DiscordVoiceState

__all__ = ("Player",)


logger: logging.Logger = logging.getLogger(__name__)


VoiceChannel = Union[
    discord.VoiceChannel, discord.StageChannel
]  # todo: VocalGuildChannel?


class Player(discord.VoiceProtocol):
    """Wavelink Player class.

    This class is used as a :class:`discord.VoiceProtocol` and inherits all its members.


    .. note::

        The Player class come with an in-built queue. See :class:`queue.Queue`.

    Parameters
    ----------
    nodes: Optional[list[:class:`node.Node`]]
        An optional list of :class:`node.Node` to use with this Player. If no Nodes are provided
        the best connected Node will be used.
    swap_node_on_disconnect: bool
        If a list of :class:`node.Node` is provided the Player will swap Nodes on Node disconnect.
        Defaults to True.

    Attributes
    ----------
    client: :class:discord.Client`
        The discord Client or Bot associated with this Player.
    channel: :class:`discord.VoiceChannel`
        The channel this player is currently connected to.
    nodes: list[:class:`node.Node`]
        The list of Nodes this player is currently using.
    current_node: :class:`node.Node`
        The Node this player is currently using.
    queue: :class:`queue.Queue`
        The wavelink built in Queue. See :class:`queue.Queue`. This queue always takes precedence over the auto_queue.
        Meaning any songs in this queue will be played before auto_queue songs.
    auto_queue: :class:`queue.Queue`
        The built-in AutoPlay Queue. This queue keeps track of recommended songs only.
        When a song is retrieved from this queue in the AutoPlay event, it is added to the main Queue.history.
    filter: dict[:class:`str`, :class:`Any`]
        The current filters applied.
    """

    def __call__(self, client: discord.Client, channel: VoiceChannel) -> Self:
        self.client = client
        self.channel = channel

        return self

    def __init__(
        self,
        client: discord.Client = MISSING,
        channel: VoiceChannel = MISSING,
        *,
        nodes: list[Node] | None = None,
        swap_node_on_disconnect: bool = True
    ) -> None:
        self.client: discord.Client = client
        self.channel: VoiceChannel | None = channel

        self.nodes: list[Node]
        self.current_node: Node

        if swap_node_on_disconnect and not nodes:
            self.nodes = list(NodePool.nodes.values())
            self.current_node = self.nodes[0]
        elif nodes:
            self.current_node = nodes[0]
            self.nodes = nodes
        else:
            self.current_node = NodePool.get_connected_node()
            self.nodes = [self.current_node]

        if not self.client:
            if self.current_node.client is None:
                raise RuntimeError('')
            self.client = self.current_node.client

        self._guild: discord.Guild | None = None
        self._voice_state: DiscordVoiceState = {}
        self._player_state: dict[str, Any] = {}

        self.swap_on_disconnect: bool = swap_node_on_disconnect

        self.last_update: datetime.datetime | None = None
        self.last_position: int = 0

        self._ping: int = 0

        self.queue: Queue = Queue()
        self._current: Playable | None = None

        self._volume: int = 50
        self._paused: bool = False

        self._autoplay: bool = False
        self.auto_queue: Queue = Queue()
        self._auto_threshold: int = 20
        self._filter: Filter | None = None

    async def _auto_play_event(self, payload: TrackEventPayload) -> None:
        if not self.autoplay:
            return

        if self.queue:
            populate = len(self.auto_queue) < self._auto_threshold
            await self.play(self.queue.get(), populate=populate)

            return

        if not self.auto_queue:
            return

        await self.queue.put_wait(await self.auto_queue.get_wait())
        populate = self.auto_queue.is_empty

        await self.play(await self.queue.get_wait(), populate=populate)

    @property
    def autoplay(self) -> bool:
        """Bool whether the Player is in AutoPlay mode or not.

        Can be set to True or False.
        """
        return self._autoplay

    @autoplay.setter
    def autoplay(self, value: bool) -> None:
        """Set AutoPlay to True or False."""
        self._autoplay = value

    def is_playing(self) -> bool:
        """Whether the Player is currently playing a track."""
        return self.current is not None

    def is_paused(self) -> bool:
        """Whether the Player is currently paused."""
        return self._paused

    @property
    def volume(self) -> int:
        """The current volume of the Player."""
        return self._volume

    @property
    def guild(self) -> discord.Guild | None:
        """The discord Guild associated with the Player."""
        return self._guild

    @property
    def position(self) -> float:
        """The position of the currently playing track in milliseconds."""

        if not self.is_playing():
            return 0

        if self.is_paused():
            return min(self.last_position, self.source.duration)  # type: ignore

        delta = (datetime.datetime.now(datetime.timezone.utc) - self.last_update).total_seconds() * 1000
        position = self.last_position + delta

        return min(position, self.current.duration)

    @property
    def ping(self) -> int:
        """The ping to the discord endpoint in milliseconds."""
        return self._ping

    @property
    def current(self) -> Playable | None:
        """The currently playing Track if there is one.

        Could be None if no Track is playing.
        """
        return self._current

    @property
    def filter(self) -> dict[str, Any]:
        return self._filter._payload

    async def _update_event(self, data: PlayerUpdateOp | None) -> None:
        assert self._guild is not None

        if data is None: 
            if self.swap_on_disconnect:

                if len(self.nodes) < 2:
                    return

                new: Node = [n for n in self.nodes if n != self.current_node and n.status is NodeStatus.CONNECTED][0]
                del self.current_node._players[self._guild.id]

                if not new:
                    return

                self.current_node: Node = new
                new._players[self._guild.id] = self

                await self._dispatch_voice_update()
                await self._swap_state()
            return

        data.pop('op')  # type: ignore
        self._player_state.update(**data)

        state: PlayerState = data['state']
        self.last_update = datetime.datetime.fromtimestamp(state.get("time", 0) / 1000, datetime.timezone.utc)
        self.last_position = state.get('position', 0)

        self._ping = state['ping']

    async def on_voice_server_update(self, data: VoiceServerUpdate) -> None:
        self._voice_state['token'] = data['token']
        self._voice_state['endpoint'] = data['endpoint']

        await self._dispatch_voice_update()

    async def on_voice_state_update(self, data: GuildVoiceState) -> None:
        assert self._guild is not None

        channel_id = data["channel_id"]

        if not channel_id:
            await self._destroy()
            return

        self._voice_state['session_id'] = data['session_id']
        self.channel = self.client.get_channel(int(channel_id))  # type: ignore

        if not self._guild:
            self._guild = self.channel.guild  # type: ignore
            assert self._guild is not None
            self.current_node._players[self._guild.id] = self

    async def _dispatch_voice_update(self, data: DiscordVoiceState | None = None) -> None:
        assert self._guild is not None

        data = data or self._voice_state

        try:
            session_id: str = data['session_id']
            token: str = data['token']
            endpoint: str = data['endpoint']
        except KeyError:
            return

        voice: Request = {'voice': {'sessionId': session_id, 'token': token, 'endpoint': endpoint}}
        self._player_state.update(**voice)

        resp: dict[str, Any] = await self.current_node._send(method='PATCH',
                                                             path=f'sessions/{self.current_node._session_id}/players',
                                                             guild_id=self._guild.id,
                                                             data=voice)

        logger.debug(f'Dispatching VOICE_UPDATE: {resp}')

    async def connect(self, *, timeout: float, reconnect: bool, **kwargs: Any) -> None:
        if self.channel is None:
            raise RuntimeError('')

        if not self._guild:
            self._guild = self.channel.guild
            self.current_node._players[self._guild.id] = self

        await self.channel.guild.change_voice_state(channel=self.channel, **kwargs)

    async def move_to(self, channel: discord.VoiceChannel) -> None:
        """|coro|

        Moves the player to a different voice channel.

        Parameters
        -----------
        channel: :class:`discord.VoiceChannel`
            The channel to move to. Must be a voice channel.
        """
        await self.guild.change_voice_state(channel=channel)
        logger.info(f"Moving to voice channel:: {channel.id}")

    async def play(self,
                   track: Playable,
                   replace: bool = True,
                   start: int | None = None,
                   end: int | None = None,
                   volume: int | None = None,
                   *,
                   populate: bool = False
                   ) -> Playable:
        """|coro|

        Play a WaveLink Track.

        Parameters
        ----------
        track: :class:`tracks.Playable`
            The :class:`tracks.Playable` track to start playing.
        replace: bool
            Whether this track should replace the current track. Defaults to ``True``.
        start: Optional[int]
            The position to start the track at in milliseconds.
            Defaults to ``None`` which will start the track at the beginning.
        end: Optional[int]
            The position to end the track at in milliseconds.
            Defaults to ``None`` which means it will play until the end.
        volume: Optional[int]
            Sets the volume of the player. Must be between ``0`` and ``1000``.
            Defaults to ``None`` which will not change the volume.
        populate: bool
            Whether to populate the AutoPlay queue. This is done automatically when AutoPlay is on.
            Defaults to False.

        Returns
        -------
        :class:`tracks.Playable`
            The track that is now playing.
        """
        assert self._guild is not None

        if isinstance(track, spotify.SpotifyTrack):
            track = await track.fulfill(player=self, cls=YouTubeTrack, populate=populate)

        data = {
            'encodedTrack': track.encoded,
            'position': start or 0,
            'volume': volume or self._volume
        }

        if end:
            data['endTime'] = end

        resp: dict[str, Any] = await self.current_node._send(method='PATCH',
                                                             path=f'sessions/{self.current_node._session_id}/players',
                                                             guild_id=self._guild.id,
                                                             data=data,
                                                             query=f'noReplace={not replace}')

        self._player_state['track'] = resp['track']['encoded']
        self._current = track

        return track

    async def set_volume(self, value: int) -> None:
        """|coro|

        Set the Player volume.

        Parameters
        ----------
        value: int
            A volume value between 0 and 1000.
        """
        assert self._guild is not None

        self._volume = max(min(value, 1000), 0)

        resp: dict[str, Any] = await self.current_node._send(method='PATCH',
                                                             path=f'sessions/{self.current_node._session_id}/players',
                                                             guild_id=self._guild.id,
                                                             data={'volume': self._volume})

        logger.debug(f'Player {self.guild.id} volume was set to {self._volume}.')

    async def seek(self, position: int) -> None:
        """|coro|

        Seek to the provided position, in milliseconds.

        Parameters
        ----------
        position: int
            The position to seek to in milliseconds.
        """
        if not self._current:
            return

        resp: dict[str, Any] = await self.current_node._send(method='PATCH',
                                                             path=f'sessions/{self.current_node._session_id}/players',
                                                             guild_id=self._guild.id,
                                                             data={'position': position})

        logger.debug(f'Player {self.guild.id} seeked current track to position {position}.')

    async def pause(self) -> None:
        """|coro|

        Pauses the Player.
        """
        assert self._guild is not None

        resp: dict[str, Any] = await self.current_node._send(method='PATCH',
                                                             path=f'sessions/{self.current_node._session_id}/players',
                                                             guild_id=self._guild.id,
                                                             data={'paused': True})

        self._paused = True
        logger.debug(f'Player {self.guild.id} was paused.')

    async def resume(self) -> None:
        """|coro|

        Resumes the Player.
        """
        assert self._guild is not None

        resp: dict[str, Any] = await self.current_node._send(method='PATCH',
                                                             path=f'sessions/{self.current_node._session_id}/players',
                                                             guild_id=self._guild.id,
                                                             data={'paused': False})

        self._paused = False
        logger.debug(f'Player {self.guild.id} was resumed.')

    async def stop(self) -> None:
        """|coro|

        Stops the currently playing Track.
        """
        assert self._guild is not None

        resp: dict[str, Any] = await self.current_node._send(method='PATCH',
                                                             path=f'sessions/{self.current_node._session_id}/players',
                                                             guild_id=self._guild.id,
                                                             data={'encodedTrack': None})

        self._player_state['track'] = None
        logger.debug(f'Player {self.guild.id} was stopped.')

    async def set_filter(
        self,
        _filter: Filter,
        /, *,
        seek: bool = False
    ) -> None:
        """|coro|

        Set the player's filter.

        Parameters
        ----------
        filter: :class:`wavelink.Filter`
            The filter to apply to the player.
        seek: bool
            Whether to seek the player to its current position
            which will apply the filter immediately. Defaults to ``False``.
        """

        assert self._guild is not None

        self._filter = _filter
        data: Request = {"filters": _filter._payload}

        resp: dict[str, Any] = await self.current_node._send(method='PATCH',
                                                             path=f'sessions/{self.current_node._session_id}/players',
                                                             guild_id=self._guild.id,
                                                             data=data)        

        if self.is_playing() and seek:
            await self.seek(int(self.position * 1000))
        logger.debug(f"Set filter:: {self._filter} ({self.channel.id})")

    async def _destroy(self) -> None:
        self.autoplay = False
        self._voice_state = {}
        self._player_state = {}
        self.cleanup()

        await self.current_node._send(method='DELETE',
                                      path=f'sessions/{self.current_node._session_id}/players',
                                      guild_id=self._guild.id)

        del self.current_node._players[self.guild.id]
        logger.debug(f'Player {self.guild.id} was destroyed.')

    async def disconnect(self, **kwargs) -> None:
        """|coro|

        Disconnect the Player from voice and cleanup the Player state.
        """
        await self.guild.change_voice_state(channel=None)

    async def _swap_state(self) -> None:
        assert self._guild is not None

        try:
            self._player_state['track']
        except KeyError:
            return

        data: EncodedTrackRequest = {'encodedTrack': self._player_state['track'], 'position': self.position}
        resp: dict[str, Any] = await self.current_node._send(method='PATCH',
                                                             path=f'sessions/{self.current_node._session_id}/players',
                                                             guild_id=self._guild.id,
                                                             data=data)

        logger.debug(f'Swapping State: {resp}')
